@@ -29,6 +29,7 @@ static int show_cc_inlined;
 static int show_cc_uninlined;
 static char *symtab_name;
 static bool show_prototypes;
+static bool show_all_matches;
 static bool expand_types;
 static bool compilable_output;
 static struct type_emissions emissions;
@@ -42,6 +43,8 @@ static struct conf_load conf_load = {
 	.conf_fprintf = &conf,
 };
 
+static struct languages languages;
+
 struct fn_stats {
 	struct list_head node;
 	struct tag	 *tag;
@@ -49,6 +52,7 @@ struct fn_stats {
 	uint32_t	 nr_expansions;
 	uint32_t	 size_expansions;
 	uint32_t	 nr_files;
+	bool		 printed;
 };
 
 static struct fn_stats *fn_stats__new(struct tag *tag, const struct cu *cu)
@@ -95,11 +99,13 @@ static void fn_stats__delete_list(void)
 	}
 }
 
-static void fn_stats__add(struct tag *tag, const struct cu *cu)
+static struct fn_stats *fn_stats__add(struct tag *tag, const struct cu *cu)
 {
 	struct fn_stats *fns = fn_stats__new(tag, cu);
 	if (fns != NULL)
 		list_add(&fns->node, &fn_stats__list);
+
+	return fns;
 }
 
 static void fn_stats_inline_exps_fmtr(const struct fn_stats *stats)
@@ -365,12 +371,19 @@ do_parameters:
 static void function__show(struct function *func, struct cu *cu)
 {
 	struct tag *tag = function__tag(func);
+	struct fn_stats *fstats = NULL;
 
-	if (func->abstract_origin || func->external)
+	if (func->abstract_origin || func->declaration)
 		return;
 
-	if (expand_types)
-		function__emit_type_definitions(func, cu, stdout);
+	if (!show_all_matches) {
+		fstats = fn_stats__find(func->name);
+
+		if (fstats && fstats->printed)
+			return;
+		if (expand_types)
+			function__emit_type_definitions(func, cu, stdout);
+	}
 	tag__fprintf(tag, cu, &conf, stdout);
 	if (compilable_output) {
 		struct tag *type = cu__type(cu, func->proto.tag.type);
@@ -391,8 +404,15 @@ static void function__show(struct function *func, struct cu *cu)
 		fprintf(stdout, "\n}\n");
 	}
 	putchar('\n');
+	if (show_all_matches)
+		return;
 	if (show_variables || show_inline_expansions)
 		function__fprintf_stats(tag, cu, &conf, stdout);
+
+	if (!fstats)
+		fstats = fn_stats__add(tag, cu);
+	if (fstats)
+		fstats->printed = true;
 }
 
 static int cu_function_iterator(struct cu *cu, void *cookie)
@@ -404,8 +424,6 @@ static int cu_function_iterator(struct cu *cu, void *cookie)
 		if (cookie && strcmp(function__name(function), cookie) != 0)
 			continue;
 		function__show(function, cu);
-		if (!expand_types)
-			return 1;
 	}
 	return 0;
 }
@@ -501,15 +519,27 @@ static enum load_steal_kind pfunct_stealer(struct cu *cu,
 
 		if (tag) {
 			function__show(tag__function(tag), cu);
-			return LSK__STOP_LOADING;
+			return show_all_matches ? LSK__DELETE : LSK__STOP_LOADING;
 		}
 	} else if (class_name) {
 		cu_class_iterator(cu, class_name);
+	} else if (show_all_matches) {
+		struct function *pos;
+		uint32_t id;
+
+		cu__for_each_function(cu, id, pos)
+			function__show(pos, cu);
 	}
 
+
 	return LSK__DELETE;
 }
 
+static struct cu *cu__filter(struct cu *cu)
+{
+	return languages__cu_filtered(&languages, cu, verbose) ? NULL : cu;
+}
+
 /* Name and version of program.  */
 ARGP_PROGRAM_VERSION_HOOK_DEF = dwarves_print_version;
 
@@ -524,6 +554,11 @@ static const struct argp_option pfunct__options[] = {
 		.arg  = "ADDR",
 		.doc  = "show just the function that where ADDR is",
 	},
+	{
+		.key  = 'A',
+		.name = "all",
+		.doc  = "show all functions that match filter, or show all function prototypes if no filter is specified",
+	},
 	{
 		.key  = 'b',
 		.name = "expand_types",
@@ -659,6 +694,7 @@ static error_t pfunct__options_parser(int key, char *arg,
 		break;
 	case 'a': addr = strtoull(arg, NULL, 0);
 		  conf_load.get_addr_info = true;	 break;
+	case 'A': show_all_matches = true;		 break;
 	case 'b': expand_types = true;
 		  type_emissions__init(&emissions, &conf);	 break;
 	case 'c': class_name = arg;			 break;
@@ -722,6 +758,12 @@ int main(int argc, char *argv[])
                 goto out;
 	}
 
+	if (languages__init(&languages, argv[0]))
+		return rc;
+
+	if (languages.exclude)
+		conf_load.early_cu_filter = cu__filter;
+
 	if (symtab_name != NULL)
 		return elf_symtabs__show(argv + remaining);
 
@@ -738,7 +780,7 @@ int main(int argc, char *argv[])
 		goto out_dwarves_exit;
 	}
 
-	if (function_name || class_name)
+	if (function_name || (!function_name && show_all_matches) || class_name)
 		conf_load.steal = pfunct_stealer;
 
 try_sole_arg_as_function_name:
